# platforms.py
import typing


@typing.no_type_check
def get_methods():
    """
    Return a dictionary with methods to deploy and test functions for each
    platform (cdf or local). This decouples platforms and deployment

    Returns:
        (dict): dictionary with methods for deploying and testing functions for each platform supported by the framework
    """
    methods = {}

    # cdf
    from akerbp.mlops.cdf.helpers import (
        cdf_metadata_validation as cdf_metadata_validation,
    )
    from akerbp.mlops.cdf.helpers import deploy_function as cdf_deploy
    from akerbp.mlops.cdf.helpers import redeploy_function as cdf_redeploy
    from akerbp.mlops.cdf.helpers import test_function as cdf_test

    methods["cdf"] = cdf_deploy, cdf_redeploy, cdf_test, cdf_metadata_validation

    # local (methods don't do anything)
    local_deploy = local_test = local_redeploy = local_metadata_validation = (
        lambda *args, **kwargs: None
    )
    methods["local"] = (
        local_deploy,
        local_redeploy,
        local_test,
        local_metadata_validation,
    )

    return methods



"""
service.py

Prediction service.
"""

import ast
import json
import traceback
import uuid
from datetime import datetime
from importlib import import_module
from typing import Dict

from cognite.client import CogniteClient
from cognite.client.exceptions import CogniteNotFoundError

from akerbp.mlops.core import config
from akerbp.mlops.core.exceptions import MissingFieldError
from akerbp.mlops.core.helpers import input_data_validation, response_validation
from akerbp.mlops.core.logger import get_logger

c = config.read_service_settings()
model_module = import_module(c.model_import_path)
predict = model_module.predict
_initialization = model_module.initialization
ModelException = model_module.ModelException

logger = get_logger(__name__)


def initialization(secrets: Dict) -> None:
    """
    Read initialization object required by `predict`
    """
    # This check adds a little overhead to each prediction
    if "init_object" not in globals():
        global init_object
        artifact_folder = c.artifact_folder
        init_object = _initialization(artifact_folder, secrets)  # type: ignore


def parse_payload(input_data: Dict, client: CogniteClient) -> Dict:
    input_wells = input_data["input"]

    for i, well_data in enumerate(input_wells):
        if input_file_reference := well_data.get("input_file_reference", ""):
            logger.info(
                f"Received input file reference: {input_file_reference}. Downloading input logs from CDF"
            )
            try:
                start = datetime.now()
                bytes_data = client.files.download_bytes(
                    external_id=input_file_reference
                )
                input_logs = json.loads(bytes_data)
                input_data["input"][i]["input_logs"] = input_logs
                logger.info(
                    f"Downloading input logs from file reference took {datetime.now() - start} s!"
                )
            except Exception as e:
                raise Exception(
                    f"Failed to load input file from CDF for file reference {well_data['input_file_reference']}"
                ) from e
        else:
            logger.warning(
                "DeprecationWarning: Sending input logs directly to the model will soon be deprecated. Migrate to sending a reference to input files in CDF using the 'input_file_reference' field to suppress this warning",
            )

    return input_data


def service(data: Dict, secrets: Dict, platform: str = "cdf") -> Dict:
    """
    Generate prediction for an input
    If the input dictionary (data) contains a key-value pair "return_file" = True,
    the resulting predictions are uploaded to Files in CDF.
    The response will now contain a field 'prediction_file' with a reference to a binary file
    containing the predictions. Otherwise, i.e. if "return_file" = False or the input dictionary does
    not contain a "return_file" key, the predictions are passed to the 'prediction' field of the response,
    and the field 'prediction_file' is set to False.

    Inputs:
        data: input to the model (sent by a user through the API)
        secrets: api keys that the model may need during initialization
    Output:
        Dictionary containing the function call response with a status field ('ok' or 'error').
        If status is 'ok' the response will have fields for 'prediction' and 'prediction_file'
        Otherwise, the response contains a field 'message' with the corresponding error message
    """
    completed_successfully = True
    try:
        if platform == "cdf":
            import akerbp.mlops.cdf.helpers as mlops_helpers

            mlops_helpers.client_secrets = secrets
            mlops_helpers.set_up_cdf_client(context="write")
        logger.info("Initializing model")
        initialization(secrets)
        logger.info("Model initialized")

        data = parse_payload(
            input_data=data, client=mlops_helpers.global_client["read"]
        )

        try:
            skip_input_validation = ast.literal_eval(
                str(c.info["prediction"]["metadata"]["supports_external_retrieval"])
            )
        except KeyError as e:
            raise MissingFieldError(
                "Field 'supports_external_retrieval' is missing from the metadata specification in mlops_settings.yaml"
            ) from e
        if skip_input_validation:
            logger.info(
                "Skipping input data validation as data is retrieved externally"
            )
            logger.info(
                "Call predict on the initialized model using the retrieved data"
            )
        else:
            logger.info("Performing input data validation")
            required_input = ast.literal_eval(
                str(c.info["prediction"]["metadata"]["required_input"])
            )
            is_input_data_valid = input_data_validation(
                required_input=required_input,
                input_payload=data,
            )
            if is_input_data_valid:
                logger.info("Input data successfully validated")
            else:
                raise KeyError(
                    f"Payload is missing at least one of the required curves: {required_input}"
                )
            logger.info(
                "Call predict on the initialized model using the provided payload"
            )
        y = predict(data, init_object, secrets)  # type: ignore
        logger.info("Performing response validation")
        is_response_valid = response_validation(response=y)
        if is_response_valid:
            logger.info("Response succesfully validated")
        else:
            raise Exception(
                "Response structure is invalid. See above warning for the set of required fields in the response"
            )

        logger.info("Predictions obtained")
        write_predictions_to_file = data.get("return_file", False)
        if write_predictions_to_file:
            logger.info("Writing predictions to file")
            if platform == "cdf":
                import base64

                public_key = data.get("public_key", None)
                content = str(json.dumps(y))
                data_encrypted = False
                if public_key is None:
                    logger.warning(
                        "Public key is not provided. The predictions will not be encrypted. This is not recommended!"
                    )
                    encrypted_content = None
                    encrypted_key = None
                else:
                    import os

                    from cryptography.hazmat.primitives import hashes, padding
                    from cryptography.hazmat.primitives.asymmetric import (
                        padding as asym_padding,
                    )
                    from cryptography.hazmat.primitives.ciphers import (
                        Cipher,
                        algorithms,
                        modes,
                    )
                    from cryptography.hazmat.primitives.serialization import (
                        load_pem_public_key,
                    )

                    logger.info("Encrypting predictions")
                    # Encrypt content with AES
                    key = os.urandom(32)
                    iv = b"1234567890123456"
                    algo = algorithms.AES(key)
                    encryptor = Cipher(algo, modes.CBC(iv)).encryptor()
                    padder = padding.PKCS7(algo.block_size).padder()
                    padded_data = padder.update(content.encode()) + padder.finalize()
                    encrypted_content = (
                        encryptor.update(padded_data) + encryptor.finalize()
                    )
                    logger.info("Encrypting AES key with RSA public key")
                    # Encrypt key with RSA public key
                    public_key = load_pem_public_key(
                        public_key.encode(),
                    )
                    encrypted_key = public_key.encrypt(
                        key,
                        asym_padding.OAEP(
                            mgf=asym_padding.MGF1(algorithm=hashes.SHA256()),
                            algorithm=hashes.SHA256(),
                            label=None,
                        ),
                    )
                    data_encrypted = True
                    logger.info("Predictions encrypted")
                cdf_client = mlops_helpers.global_client["write"]
                prediction_id = uuid.uuid4().hex
                external_file_id = f"{c.model_name}_predictions_{prediction_id}.binary"
                try:
                    cdf_client.files.delete(external_id=external_file_id)
                except CogniteNotFoundError:
                    pass
                cdf_client.files.upload_bytes(
                    content=encrypted_content
                    if encrypted_content is not None
                    else content,
                    name=f"{c.model_name}_predictions_{prediction_id}",
                    external_id=external_file_id,
                )
                logger.info(f"Prediction file uploaded to {external_file_id}")
                return {
                    "status": "ok",
                    "prediction": {},
                    "prediction_file": external_file_id,
                    "model_id": c.model_id,
                    "data_encrypted": data_encrypted,
                    "encrypted_key": base64.b64encode(encrypted_key).decode()
                    if encrypted_key is not None
                    else None,
                }
            elif platform == "gc":
                raise NotImplementedError
            else:
                raise ValueError(f"Platform {platform} is not supported")
        else:
            logger.info("Writing predictions to the response")
            return {
                "status": "ok",
                "prediction": y,
                "prediction_file": "",
                "model_id": c.model_id,
            }
    except KeyError:
        completed_successfully = False
        error_message = "Unable to obtain a prediction with the provided payload"
    except ModelException as error:
        completed_successfully = False
        error_message = f"Could not get a prediction. Message: {error}"
    finally:
        if not completed_successfully:
            error_message += (
                f" See the traceback for more details:\n{traceback.format_exc()}"
            )
            logger.error(error_message)
    return {"status": "error", "message": error_message, "model_id": c.model_id}



# handler.py

import sys
import time
import traceback
from contextlib import redirect_stderr
from importlib import import_module
from typing import Any, Dict, Union


def handle(
    data: Dict, secrets: Dict, function_call_info: Dict
) -> Union[Any, Dict[Any, Any]]:
    """Handler function for deploying models to CDF

    Args:
        data (Dict): model payload
        secrets (Dict): Client secrets to be used by the service
        function_call_info (Dict): dictionary containing function id and whether the function call is scheduled

    Returns:
        Union[Any, Dict[Any, Any]]: Function call response
    """
    try:
        with redirect_stderr(sys.stdout):
            import warnings

            import akerbp.mlops.cdf.helpers as cdf
            from akerbp.mlops.core.config import ENV_VARS
            from akerbp.mlops.core.logger import get_logger

            def customwarn(message, category, filename, lineno, file=None, line=None):
                sys.stdout.write(
                    warnings.formatwarning(message, category, filename, lineno)
                )

            warnings.showwarning = customwarn

            service_name = ENV_VARS.service_name
            service = import_module(f"akerbp.mlops.services.{service_name}").service

            logger = get_logger(__name__)
            cdf.client_secrets = secrets
            logger.info("Prediction request received, starting model prediction")
            logger.info(
                "Setting up CDF Client with access to Data, Files and Functions"
            )
            cdf.set_up_cdf_client(context="read")
            logger.info("Set up complete")
            if data:
                logger.info("Calling model using provided payload")
                start = time.time()
                output = service(data, secrets)
                elapsed = time.time() - start
                logger.info(f"Model call complete. Duration: {elapsed:.2f} s")
            else:
                logger.info("Calling model with empty payload")
                output = {"status": "ok"}
                logger.info("Model call complete")
            logger.info("Querying metadata from the function call")
            function_call_metadata = cdf.get_function_call_response_metadata(
                function_call_info["function_id"]
            )
            logger.info("Function call metadata obtained")
            logger.info("Writing function call metadata to response")
            output.update({"metadata": function_call_metadata})
            logger.info("Function call metadata successfully written to response")
            return output
    except Exception:
        trace = traceback.format_exc()
        error_message = ""
        try:
            error_message = f"{service_name} service failed.\n{trace}"
            logger.critical(error_message)
        except UnboundLocalError:
            if not error_message:
                error_message = "Service failed. \n{trace}"
            pass  # Unable to log error just return error message
        return {"status": "error", "error_message": error_message}



import argparse
import os

import akerbp.mlops.model_manager as mm


def main():
    parser = argparse.ArgumentParser(
        description="Promote artifacts to specified target environment"
    )
    parser.add_argument(
        "-n",
        "--model_name",
        type=str,
        help="Name of the model to promote. If not provided, it is inferred from the mlops_settings.yaml file.",
        required=False,
    )
    args = parser.parse_args()
    model_name = args.model_name
    if model_name is None:
        from akerbp.mlops.core.config import read_project_settings

        model_names = [s.model_name for s in read_project_settings()]
    else:
        model_names = [model_name]
    env = os.environ.get("MODEL_ENV", None)
    if env is None:
        raise Exception("Environment variabel 'MODEL_ENV' is not set")

    if env == "dev":
        target_env = "test"
    elif env == "test":
        target_env = "prod"
    else:
        raise Exception(
            "Cannot infer target env from environment variable 'MODEL_ENV'={env}"
        )

    mm.setup()
    for model_name in model_names:
        mm.promote_model(
            model_name=model_name,
            promote_to=target_env,
            confirm=False,
        )


if __name__ == "__main__":
    main()  # type: ignore



import argparse
import logging
import os
import subprocess

import akerbp.mlops.model_manager as mm


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-n",
        "--model_name",
        type=str,
        help="Name of model. If not provided, it will be inferred from the mlops_settings.yaml file.",
        default="",
        required=False,
    )
    args = parser.parse_args()
    model_name = args.model_name
    if not model_name:
        from akerbp.mlops.core.config import read_project_settings

        model_names = [s.model_name for s in read_project_settings()]
    else:
        model_names = [model_name]

    logging.disable()
    mm.setup()
    env = os.environ.get("MODEL_ENV")
    try:
        commit = subprocess.check_output(
            ["git", "rev-parse", "--short", "HEAD"], text=True, stderr=subprocess.STDOUT
        ).strip()
    except subprocess.CalledProcessError as e:
        raise Exception(f"Error getting commit hash to tag: {e}") from e

    for model_name in model_names:
        try:
            latest_deployed_model = (
                mm.get_model_version_overview(
                    model_name=model_name,
                    env=env,
                    output_logs=False,
                )
                .sort_values(by="uploaded_time", ascending=False)
                .iloc[0]
            )
            version_number = latest_deployed_model.external_id.split("/")[-1]
        except IndexError as e:
            raise Exception(
                f"No version of model {model_name} found in the model registry"
            ) from e

        tag = f"v{version_number}_{model_name}_{env}"
        # Delete tag if it exists
        try:
            subprocess.run(
                ["git", "tag", "-d", tag],
                check=True,
                text=True,
                stderr=subprocess.STDOUT,
            )
            subprocess.run(
                ["git", "push", "origin", f":refs/tags/{tag}"],
                check=True,
                text=True,
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            logging.warning(f"Error deleting tag: {e}")

        # Create tag
        try:
            subprocess.run(
                ["git", "tag", tag, commit],
                check=True,
                text=True,
                stderr=subprocess.STDOUT,
            )
            subprocess.run(
                ["git", "push", "origin", tag],
                check=True,
                text=True,
                stderr=subprocess.STDOUT,
            )
        except subprocess.CalledProcessError as e:
            raise Exception(f"Error creating tag: {e}") from e


if __name__ == "__main__":
    main()  # type: ignore



# setup.py
import re
from copy import deepcopy
from pathlib import Path

from akerbp.mlops import __version__ as version
from akerbp.mlops.core.config import (
    ENV_VARS,
    generate_default_project_settings,
    validate_user_settings,
)
from akerbp.mlops.core.logger import get_logger
from akerbp.mlops.deployment.helpers import replace_string_file, to_folder

logger = get_logger(__name__)


def update_mlops_version_in_pipeline(pipeline_filepath) -> None:
    """Update MLOps settings by overwriting the package version number.
    First, it will go through the pipeline definition and remove any quoatation mark prefix to akerbp.mlops
    (to make it robust against updating from previous versions),
    before it will replace the version number with whatever version the caller has on his/her local machine.

    Args:
        folder_path (Path, optional): Path to folder containing the pipeline file. Defaults to Path(".").
    """
    platform = ENV_VARS.platform
    pattern_package = re.compile(r"mlops_version: \d+\.\d+\.\d+(?:(a|-alpha\.)\d+)?")
    replacement_package = f"mlops_version: {version}"
    pattern_pip_install = re.compile(
        r"akerbp-mlops(?:\[(cdf|gcp)\])?==\d+\.\d+\.\d+(?:(a|-alpha\.)\d+)?"
    )
    replacement_pip_install = f"akerbp-mlops[{platform}]=={version}"
    pattern_version = re.compile(r"Version \S{5,}")
    replacement_version = f"Version {version}"
    with pipeline_filepath.open() as f:
        pipeline = f.read()
        new_pipeline = re.sub(
            pattern_version,
            replacement_version,
            re.sub(
                pattern_package,
                replacement_package,
                re.sub(
                    pattern_pip_install,
                    replacement_pip_install,
                    pipeline,
                ),
            ),
        )
    with pipeline_filepath.open("w") as f:
        f.write(new_pipeline)
    if new_pipeline == pipeline:
        logger.warning("No changes to the MLOps version were made in the pipeline file")
    else:
        logger.info("MLOps version successfully updated in pipeline file")


def remove_references_to_old_bash_script(pipeline_filepath) -> None:
    """Remove references to the old bash script in the pipeline file

    Args:
        folder_path (Path, optional): Path to folder containing the pipeline file. Defaults to Path(".").
    """
    replacement_pattern = re.compile("bash deploy_.+.sh")
    with pipeline_filepath.open() as f:
        pipeline = f.read()
        matches = list(re.finditer(replacement_pattern, pipeline))
        if matches:
            new_pipeline = deepcopy(pipeline)
            for match in matches:
                logger.info(f"Removing reference to old bash script {match}")
                match_string = match.group()
                script_name = (
                    match_string.replace("bash", "").replace(".sh", "").strip()
                )
                new_pipeline = new_pipeline.replace(match_string, script_name)
            with pipeline_filepath.open("w") as f:
                f.write(new_pipeline)
            logger.info(
                "References to old bash script successfully removed from pipeline file"
            )
        else:
            logger.info("No references to old bash scripts found")


def generate_pipeline(pipeline_path: Path) -> None:
    """Generate pipeline definition from template

    Args:
        folder_path (Path, optional): Path to folder that should contain the pipeline file. Defaults to Path(".").
    """
    python_version = input(
        "Which python version should be used for the deployment (3.9, 3.10, 3.11, 3.12)?: "
    ).strip()
    pipeline_path.parent.mkdir(parents=True, exist_ok=True)
    logger.info(f"Moving main pipeline definition file to {pipeline_path.parent}")
    to_folder(("akerbp.mlops.deployment", "main.yml"), pipeline_path.parent)
    logger.info(
        f"Moving helper action definition files to {pipeline_path.parent.parent}"
    )
    to_folder(("akerbp.mlops.deployment", "actions"), pipeline_path.parent.parent)
    replace_string_file("MLOPS_VERSION_GOES_HERE", version, pipeline_path)
    replace_string_file(
        "PYTHON_VERSION_GOES_HERE",
        f'"{python_version}"',
        pipeline_path.parent.parent
        / "actions/deploy-prediction-service-action/action.yml",
    )
    logger.info("Pipeline definition generated")


def setup_pipeline(folder_path: Path) -> None:
    """
    Set up pipeline file in the given folder.
    Update MLOps package version in the pipeline file if it exists, or generate from template if it doesn't.

    Args:
        folder_path (Path, optional): Path to folder that should contain the pipeline file. Defaults to Path(".").
    """
    pipeline_path = folder_path.resolve() / ".github" / "workflows" / "main.yml"
    if pipeline_path.exists():
        logger.info(f"Update MLOps version in pipeline definition to {version}")
        update_mlops_version_in_pipeline(pipeline_path)
        remove_references_to_old_bash_script(pipeline_path)
    else:
        logger.info(
            "No existing deployment workflow found. Generating pipeline definition from template"
        )
        generate_pipeline(pipeline_path)


if __name__ == "__main__":
    logger = get_logger(__name__)

    default_path = Path("mlops_settings.yaml").resolve()
    if default_path.exists():
        logger.info("Validate settings file")
        validate_user_settings(default_path)
    else:
        logger.info("Create settings file template")
        generate_default_project_settings(default_path, n_models=1)
    setup_pipeline(default_path.parent)
    logger.info("Done!")



"""Model Manager module for handling model artifacts in CDF.

This module deals with uploading, promoting and downloading of model artifacts
to/from CDF.
"""

from pathlib import Path
from shutil import unpack_archive
from typing import Any, Dict, Optional, Union

import pandas as pd
from cognite.client.exceptions import CogniteAPIError

import akerbp.mlops.cdf.helpers as cdf
from akerbp.mlops.core import config
from akerbp.mlops.core.exceptions import (
    MissingClientError,
    MissingDeploymentFolder,
    MissingModelArtifactsError,
    MissingSettingsFileError,
)
from akerbp.mlops.core.helpers import confirm_prompt
from akerbp.mlops.core.logger import get_logger

logger = get_logger(__name__)

env = config.ENV_VARS.model_env
client_secrets = config.client_secrets
dataset_id = "mlops"


def setup(
    client_secrets: Optional[Dict] = None, dataset_external_id: str = "mlops"
) -> None:
    """
    Set up the model manager. This involves setting up the CDF client and the
    dataset used to store artifacts.

    Args:
        client_secrets (:obj:`dict`, optional): dictionary with cdf client secrets
        dataset_external_id (str): external id for the dataset (use None for no dataset)
    """
    if client_secrets:
        cdf.client_secrets = client_secrets
    try:
        cdf.set_up_cdf_client("write")
    except KeyError:
        logger.warning(
            "No client secrets found with write privileges. You can still use model_manager, but with limited (read) functionality."
        )
        cdf.set_up_cdf_client()
    set_active_dataset(dataset_external_id)


def set_active_dataset(external_id: str) -> None:
    """
    Set current active dataset

    Args:
        external_id (str): external id for the dataset (use None for no dataset)
    """
    global dataset_id
    dataset_id = cdf.get_dataset_id(external_id)
    m = f"Active dataset: {external_id=}, {dataset_id=}"
    logger.info(m)


def upload_new_model_version(
    model_name: str,
    env: str,
    folder: Path,
    metadata: Optional[Dict] = None,
    **kwargs,
) -> Any:
    """
    Upload a new model version. Files in a folder are archived and stored
    with external id `model_name/env/version`, where version is automatically
    increased.

    Args:
        model_name: name of the model
        env: name of the environment ('dev', 'test', 'prod')
        folder: (Path) path to folder whose content will be uploaded
        metadata: Optional[dictionary with metadata (it should not contain a 'version' key)

    Returns:
        (dict): model metadata
    """
    if metadata is None:
        metadata = {}
    specified_version = kwargs.get("version", None)
    if env == "prod" or env == "test":
        raise ValueError(
            f"You are not allowed to upload artifacts directly to {env}, use promote_model instead to promote artifacts dev to test and henceforth from test to prod"
        )
    if specified_version is None:
        file_list = cdf.query_file_versions(
            external_id_prefix=f"{model_name}/{env}/",
            directory_prefix="/mlops",
            uploaded=None,  # count any file
            dataset_id=dataset_id,
        )
        if not file_list.empty:
            latest_v = file_list.metadata.apply(lambda d: int(d["version"])).max()
        else:
            latest_v = 0

        version = int(latest_v) + 1  # int64 isn't json-serializable
    else:
        version = int(specified_version)

    if "version" in metadata:
        logger.error(
            "Metadata should not contain a 'version' key. " "It will be overwritten"
        )
    metadata["version"] = version
    external_id = f"{model_name}/{env}/{version}"

    if not isinstance(folder, Path):
        folder = Path(folder)

    folder_info = cdf.upload_folder(
        external_id=external_id,
        path=folder,
        metadata=metadata,
        target_folder="/mlops",
        dataset_id=dataset_id,
    )
    logger.info(
        f"Artifacts with {external_id=} uploaded from folder '{folder}' to {folder_info.directory} ({folder_info.name})"
    )
    return folder_info


def find_model_version(model_name: str, env: str, metadata: Optional[Dict]) -> str:
    """
    Model external id is specified by the model name and the environment
    (starts with `{model_name}/{env}`), and a query to the metadata. If this is
    not enough, the latest version is chosen.

    Args:
        model_name (str): name of the model
        env (str): name of the environment ('dev', 'test', 'prod')
        metadata (dict): metadata of the model artifacts
    Returns:
        (str): external id of the model
    """
    try:
        project_settings = config.read_project_settings()
    except FileNotFoundError as e:
        logger.error("No 'mlops_settings.yaml' file found.")
        raise MissingSettingsFileError("No 'mlops_settings.yaml' file found.") from e
    file_list = cdf.query_file_versions(
        directory_prefix="/mlops",
        external_id_prefix=f"{model_name}/{env}",
        metadata=metadata,
        dataset_id=dataset_id,
    )

    latest_artifact_version = None
    if (n_models := file_list.shape[0]) == 0:
        message = f"No model artifacts found for model with {model_name=}, {env=} and metadata {metadata}. \
            Upload/promote artifacts or specify the correct model name before redeploying"
        raise Exception(message)
    elif n_models > 1:
        logger.info(
            f"Found {n_models} model artifact folders in {env} environment. Choosing the latest version, or the version specified in mlops_settings.yaml"
        )
        latest_artifact_version = file_list.loc[
            file_list["uploaded_time"].argmax(), "external_id"
        ].split("/")[-1]
        logger.info(f"Latest artifact version found is {latest_artifact_version}")
    if project_settings is None:
        raise FileNotFoundError(
            "Missing 'mlops_settings.yaml' file, not possible to use this functionality"
        )

    for c in project_settings:
        if c.model_name == model_name:
            if c.artifact_version is not None:
                artifact_version = c.artifact_version
                if latest_artifact_version is not None and int(artifact_version) > int(
                    latest_artifact_version
                ):
                    raise Exception(
                        f"Artifact version {artifact_version} is greater than the number of artifacts {n_models} in {env}. Please specify a valid artifact version"
                    )
                external_id = f"{model_name}/{env}/{artifact_version}"
                logger.info(
                    f"Retrieving model with specified artifact version ({artifact_version}) in {env} environment"
                )
            else:
                external_id = str(
                    file_list.loc[file_list["uploaded_time"].argmax(), "external_id"]
                )
                artifact_version = external_id.split("/")[-1]
                logger.info(
                    f"No artifact version specified in the settings, retrieving model with latest artifact version ({artifact_version}) in {env} environment"
                )
        else:
            latest_version = str(
                file_list.loc[file_list["uploaded_time"].argmax(), "external_id"].split(
                    "/"
                )[-1]
            )
            external_id = f"{model_name}/{env}/{latest_version}"
            continue

    return external_id


def download_model_version(
    model_name: str,
    env: str,
    folder: Union[Path, str],
    metadata: Optional[Dict] = None,
    version: Optional[str] = None,
) -> str:
    """
    Download a model version to a folder. First the model's external id is found
    (unless provided by the user), and then it is downloaded to the chosen
    folder (creating the folder if necessary).

    Args:
        model_name (str): name of the model
        env (str): name of the environment ('dev', 'test', 'prod')
        folder (Union[Path,str]): path to folder where the artifacts are downloaded
        metadata (dict): metadata of the artifacts, doesn't make sense when passing a version (see next parameter)
        version (int, optional): artifact version to download from CDF
    Returns:
        (str): external id of the downloaded model artifacts
    """
    if metadata is None:
        metadata = {}
    if isinstance(folder, str):
        folder = Path(folder)

    if version:
        external_id = f"{model_name}/{env}/{version}"
    else:
        external_id = find_model_version(model_name, env, metadata)

    if not folder.exists():
        folder.mkdir()
    cdf.download_folder(external_id, folder)
    logger.info(f"Downloaded model with {external_id=} to {folder}")

    return external_id


def set_up_model_artifact(artifact_folder: Path, model_name: str) -> str:
    """
    Set up model artifacts.
    When the prediction service is deployed, we need the model artifacts. These
    are downloaded, unless there's already a folder (local development
    environment only)

    Args:
        artifact_folder (Path):
        model_name: str

    Returns:
        (str): model id provided by the model manager or an existing folder in dev
    """
    message = f"Existing, local artifacts won't be used ({env=}). Download uploaded artifacts from CDF Files"
    logger.info(message)

    logger.info("Downloading serialized model")
    model_id = download_model_version(model_name, env, artifact_folder)
    return model_id


def get_model_version_overview(
    model_name: Optional[str] = None,
    env: Optional[str] = None,
    output_logs: bool = True,
    metadata: Optional[Dict] = None,
) -> pd.DataFrame:
    """
    Get overview of model artifact versions.

    Args:
        model_name (:obj:`str`, optional): name of the model or None for any
        env (:obj:`str`, optional): name of the environment ('dev', 'test', 'prod')
        metadata (dict): artifact metadata. Defaults to an empty dictionary.

    Returns:
        (pd.DataFrame): model artifact data (external id, id, etc.)
    """
    if metadata is None:
        metadata = {}
    if env is not None:
        env = env.lower()
    # All mlops files with right metadata
    file_list = cdf.query_file_versions(
        directory_prefix="/mlops",
        external_id_prefix=None,
        uploaded=None,
        metadata=metadata,
        dataset_id=dataset_id,
    )

    # query CDF based on the external id
    if model_name:
        index = file_list.external_id.str.contains(model_name + "/")
        file_list = file_list.loc[index]
    if env:
        index = file_list.external_id.str.contains("/" + env + "/")
        file_list = file_list.loc[index]
    if not dataset_id:
        index = file_list.dataset_id.isna()
        file_list = file_list.loc[index]
    sorted_file_list = file_list.sort_values(
        by="last_updated_time",
        ascending=False,
    )
    if output_logs:
        logger.info("Sorting artifacts chronologically")
        if env is None:
            logger.info("For more granular control you can specify the 'env' argument")
    return sorted_file_list


def validate_model_id(external_id: str, verbose: bool = True) -> bool:
    """
    Validate that model id follows MLOps standard: model/env/id

    Args:
        external_id (str): model id
        verbose (bool): Whether to print a warning if model id is invalid. Defaults to True.
    Returns:
        (bool): True if name is valid, False otherwise
    """
    supported_environments = ["dev", "test", "prod"]
    try:
        _, environment, version = external_id.split("/")
    except ValueError:
        if verbose:
            m = "Expected model id format: 'model/env/id'"
            logger.error(m)
        return False
    if environment not in supported_environments:
        if verbose:
            m = f"Supported environments: {supported_environments}"
            logger.error(m)
        return False
    try:
        int(version)
    except ValueError:
        if verbose:
            m = f"Version should be integer, got '{version}' instead"
            logger.error(m)
        return False
    return True


def delete_model_version(external_id: str, confirm: bool = True) -> None:
    """
    Delete a model artifact version

    Args:
        external_id (str): artifact's external id in CDF Files.
        confirm (bool): whether the user will be asked to confirm deletion. Defaults to True.
    """
    if not validate_model_id(external_id):
        raise ValueError(
            f"Invalid specification of external id ({external_id}), use the format 'model_name/env/version'"
        )
    model, environment, version = external_id.split("/")
    if not cdf.file_exists(external_id, "/mlops"):
        raise MissingModelArtifactsError(
            f"Model artifacts for {external_id} not found in CDF Files"
        )

    confirmed = False
    if confirm:
        question = f"Delete {model=}, {environment=}, {version=}?"
        confirmed = confirm_prompt(question)

    if not confirm or confirmed:
        cdf.delete_file({"external_id": external_id})


def overwrite_artifacts(
    external_id: str,
    folder: Path,
    metadata: Optional[Dict] = None,
    confirm: bool = True,
) -> None:
    """Overwrite existing model artifacts with new ones.


    Args:
        external_id (str): external id of the model artifacts to overwrite in CDF Files,
            on the form "model_name/env/version".
        confirm (bool, optional): Whether to confirm overwriting the artifacts.
            Defaults to True.
    """

    if metadata is None:
        metadata = {}
    if not validate_model_id(external_id):
        raise ValueError(
            f"Invalid specification of external id ({external_id}), use the format 'model_name/env/version'"
        )
    model, environment, version = external_id.split("/")
    if not cdf.file_exists(external_id, "/mlops"):
        raise MissingModelArtifactsError(
            f"Model artifacts for {external_id} not found in CDF Files"
        )

    confirmed = False
    if confirm:
        question = f"Overwrite {model=}, {environment=}, {version=}?"
        confirmed = confirm_prompt(question)

    if not confirm or confirmed:
        if environment == "prod":
            raise Exception(
                "Not allowed to overwrite production artifacts. Overwrite artifacts in test, test your model and promote to production"
            )

        else:
            cdf.delete_file({"external_id": external_id})
            upload_new_model_version(
                model_name=model,
                env=environment,
                folder=folder,
                metadata=metadata,
                version=version,
            )


def promote_model(model_name: str, promote_to: str, **kwargs) -> None:
    """
    Promote a model version from test to prod

    Keyword Args:
        confirm (bool): Whether to confirm the promotion, defaults to True
        version (int): Version number. Defaults to None,
            where the latest version number is queried from CDF.
    Args:
        model_name (str): name of model
        promote_to (str): environment to promote to, either 'test' or 'prod'
    """
    version = kwargs.get("version", None)
    confirm = kwargs.get("confirm", True)

    if promote_to == "test":
        promote_from = "dev"
    elif promote_to == "prod":
        promote_from = "test"
    else:
        raise ValueError("promote_to must be either 'test' or 'prod'")

    if version is None:
        logger.info(
            f"No version number specified, will query CDF for the latest version uploaded to {promote_from}"
        )
        external_id_function = f"{model_name}-prediction-{promote_from}"
        version = cdf.get_latest_artifact_version(external_id=external_id_function)
        logger.info(f"Latest uploaded model version in {promote_from} is {version}")
    external_id = f"{model_name}/{promote_from}/{version}"
    if not cdf.file_exists(external_id, "/mlops", dataset_id):
        logger.warning(
            f"Model version {external_id} doesn't exist in {promote_from}, nothing to promote."
        )
        return

    confirmed = False
    if confirm:
        question = f"Promote artifacts for {model_name} from {promote_from}, {version=},  to {promote_to}?"
        confirmed = confirm_prompt(question)

    target_ext_id = f"{model_name}/{promote_to}/{version}"
    if cdf.file_exists(
        external_id=target_ext_id, directory="/mlops", dataset_id=dataset_id
    ):
        logger.info(
            f"Model version {target_ext_id} already exists in {promote_to}, nothing new to promote."
        )
        return

    if not confirm or confirmed:
        try:
            client = cdf.global_client["read"]
        except KeyError as e:
            raise MissingClientError(
                "Set up model manager by running 'mm.setup()' before promoting artifacts"
            ) from e

        old_filename = client.files.retrieve(external_id=external_id).name
        cdf.copy_file(
            external_id,
            target_ext_id,
            dataset_id=dataset_id,
            overwrite_name=True,
            name=old_filename.replace(promote_from, promote_to),
        )


def get_latest_model_in_env(model_name: str, env: str, deployed: bool = False) -> int:
    """Return the number of uploaded artifact versions for a given model in a given environment.
    If deployed is set to True, this function will try to return the version number of the latest deployed
    model in env. This requires that the deployed model has a metadata field 'version'. If not, and deployed is True,
    then version is set to -1.

    Note that the output of get_model_version_overview(args) is sorted chronologically, so the first
    element is the latest

    Args:
        model_name (str): model name as specified in settings file
        env (str): environment, in {"dev", "test", "prod"}
        deployed (bool): Whether we want to check for live models deployed to CDF
            Defaults to False, where the latest artifact version is returned

    Returns:
        int: number of models
    """
    if deployed:
        external_id_function = f"{model_name}-prediction-{env}"
        functions_client = cdf.get_client(
            client_id=client_secrets["id-write"],
            client_secret=client_secrets["secret-write"],
        )
        latest_deployed_model = functions_client.functions.retrieve(
            external_id=external_id_function
        )

        if latest_deployed_model is None:
            raise ValueError(
                f"Function with external id {external_id_function} does not exist in CDF"
            )
        try:
            version = latest_deployed_model.metadata["version"]

        except KeyError:
            logger.warning(
                f"Deployed function with external id {external_id_function} does not have a version attribute in the metadata"
            )
            version = -1
        return int(version)
    else:
        return int(
            get_model_version_overview(model_name=model_name, env=env)
            .external_id.iloc[0]
            .split("/")[-1]
        )


def download_deployment_folder(model_name: str, env: str, **kwargs) -> None:
    """Download deployment folder containing a subfolder with uploaded artifacts,
    model interface and test suite, also referred to as 'model code',  and requirements.
    If no version is specified, the function will extract and download the latest deployed model, unless
    latest deployed model does not have a 'version' field in the metadata, in which the latest uploaded
    artifact version is downloaded.

    If the 'target_path' is not specified in the kwargs,
    the resulting deployment folder is downloaded as a zip-file and unpacked as follows:

    mlops_deployment_folder/<model_name>/<env>/v<version>/*

    One can choose to ignore specific files by passing the 'files_to_ignore' kwargs
    as a list of filenames. The function will iterate through the entire tree and
    delete all files matching the elements in the list. This is particularly useful
    when wanting to call a model as a feature engineering step in another deployed model,
    and you want to speed up inference time by calling the model interface directly
    from source

    Args:
        model_name (str): model_name to download, as specified in mlops settings file.
        env (str): environment

    Keyword Args:
        version (int): version number, defaults to None (latest)
        files_to_ignore (list[str]): list of files to ignore, defaults to []
        target_path (str): relative target path of deployment folder, defaults to None
    """
    version = kwargs.get("version", None)
    files_to_ignore = kwargs.get("files_to_ignore", [])
    final_target_path = kwargs.get("target_path", None)
    if final_target_path is None:
        final_target_path = f"./mlops_deployment_folder/{model_name}/{env}/v{version}"
    latest_version = get_latest_model_in_env(
        model_name=model_name, env=env, deployed=True
    )
    if latest_version == -1:
        latest_version = get_latest_model_in_env(
            model_name=model_name, env=env, deployed=False
        )
    if version is None:
        version = latest_version
        logger.info(
            f"No version number specified, downloading the latest version ({version}) by default"
        )

    if latest_version < version:
        raise ValueError(
            f"Specified version number is greater than the latest uploaded artifacts for model {model_name} in {env}"
        )
    else:
        logger.info(
            f"Downloading (zipped) deployment folder for model {model_name} in {env}, version {version}"
        )
        zipped_target_path = Path(f"{model_name}-{env}-{version}_deployment.zip")
        if env == "prod":
            external_id = f"{model_name}-prediction-{env}-{version}"
        else:
            external_id = f"{model_name}-prediction-{env}"
        try:
            cdf.download_file(
                file_id={"id": None},
                path=zipped_target_path,
                external_id={"external_id": external_id},
            )
        except CogniteAPIError as e:
            raise MissingDeploymentFolder(
                f"Failed to download deployment folder for model with external id {external_id}"
            ) from e
        logger.info(f"Unzipping deployment folder => '{final_target_path}' ")
        unpack_archive(filename=zipped_target_path, extract_dir=final_target_path)

        # Ignore specified files
        for file_to_ignore in files_to_ignore:
            logger.info(
                f"Removing file '{file_to_ignore}' from folder '{final_target_path}'"
            )
            path_to_ignore = Path(final_target_path) / file_to_ignore
            try:
                path_to_ignore.unlink()
            except Exception as e:
                logger.warning(f"Failed to remove file '{path_to_ignore}': {e}")
                continue

        # Delete zipped folder
        zipped_target_path.unlink()
